!*************************************
! Jacobi Method for Laplace Equation
! with Dirichlet Boundary Conditions
! Convergence: biggest change in
! matrix smaller tolerance value
!************************************

#define ROWS 4096
#define COLS 4096

program laplace
  use realtime
#ifdef _OPENACC
  use openacc
  use checkResults
#endif
  implicit none
  integer, parameter :: fp_kind=kind(1.0d0)
  integer, parameter :: n=ROWS, m=COLS
  integer, parameter :: iter_max=20
  integer :: i, j, iter
  real(fp_kind), dimension (:,:), allocatable :: U, Unew
  real(fp_kind) :: tol=1.0e-6_fp_kind, err=1.0_fp_kind
  real(fp_kind) :: runtime


  allocate (U(n,m), Unew(n,m))

  ! Initialize arrays
  U    = 0.0_fp_kind
  Unew = 0.0_fp_kind
  U(:,1)    = -1.0_fp_kind
  Unew(:,1) = -1.0_fp_kind

#ifdef _OPENACC
  call initGPU()
#endif

  write(*,*) 'Solving Laplace Equation by Jacobi Method'
  write(*,'(a,i5,a,i5)') ' Matrix dim:', n, ' x', m
  write(*,*) 'Iteration: Error'
 
  runtime = GetRealTime()

  iter=0

  do while (err .gt. tol .and. iter .lt. iter_max)
    err=0.0_fp_kind


    do j=2,m-1
      do i=2,n-1
        Unew(i,j) = 0.25_fp_kind * (U(i+1,j  ) + U(i-1,j  ) + &
                                    U(i  ,j-1) + U(i  ,j+1))
        err = max(err, abs(Unew(i,j)-U(i,j)))
      end do
    end do




    do j=2,m-1
      do i=2,n-1
        U(i,j) = Unew(i,j)
      end do
    end do


  iter = iter + 1

  write(*,'(i9,a,f10.6)'), iter, ': ', err

  end do

  runtime = GetRealTime() - runtime 

!  passed=checkSolution(err, U, n, m)
!  if (passed >= 1) then
!     write(*,*) 'Error check failed.'
!  else
!     write(*,*) 'Error check passed.'
!  end if

  write(*,'(a,f10.3,a)')  ' Time Elapsed: ', runtime, ' s'

  deallocate (U,Unew)
end program laplace



#ifdef _OPENACC
subroutine initGPU()
  use openacc
  implicit none
  character(len=10) :: argv
  integer :: iargc, argc, devId, devCount

  argc = iargc()
  ! gets the device id (if specified) to run on
  devId = -1
  if (argc >= 1) then
     call getarg(1,argv)
     read(argv,'(i)') devId
     devCount = acc_get_num_devices(acc_device_nvidia)
     if (devId < 0 .OR. devId >= devCount) then
        write(*,*) "The specified device ID is not supported."
        stop
     end if
  end if
  if (devId /= -1) then
     call acc_set_device_num(devId, acc_device_nvidia)
  end if
  ! creates a context on the GPU just to
  ! exclude initialization time from computations
  call acc_init(acc_device_nvidia)

  ! print device id
  devId = acc_get_device_num(acc_device_nvidia)
  write(*,'(a,i1,a)') "Running on GPU with ID ",devId,".\n\n"

end subroutine initGPU
#endif
