
/* Example showing the use of CUFFT for fast 1D-convolution using FFT. OpenACC
   is used for the point-wise complex multiply and scale.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <cuda_runtime_api.h>

#include <cufft.h>

#ifdef _OPENACC
#include <openacc.h>
#endif
#include "cufft_acc.h"

// The filter size is assumed to be a number smaller than the signal size
#define SIGNAL_SIZE        500000
#define FILTER_KERNEL_SIZE 33


void check(cudaError_t result, char const *const func, const char *const file, int const line)
{
    if (result) {
        fprintf(stderr, "%s:%d code=%d (%s) \"%s\"\n", file, line, 
                (unsigned int)result, cudaGetErrorString(result));
    }
}

// This will output the proper CUDA error strings in the event that a CUDA host call returns an error
#define checkCudaErrors(val)           check ( (val), #val, __FILE__, __LINE__ )


// Complex data type
/*typedef struct {
    float x, y;
} Complex;
*/

// Padding function
int PadData(const Complex *, Complex **, int,
            const Complex *, Complex **, int);



// Correctness checking
extern int compareL2(const float* reference, const float* data,
              const unsigned int len, const float epsilon);
// Host filtering function
extern void Convolve(const Complex *, int, const Complex *, int, Complex *);

// main test function
void runTest(int argc, char **argv);
// initialization of GPU (device id)
void initGPU(int argc, char **argv);

////////////////////////////////////////////////////////////////////////////////
// Task: Implement and run on GPU with OpenACC
// Hint: Lookup the deviceptr clause
////////////////////////////////////////////////////////////////////////////////
void complexPointwiseMulAndScale(int n, float *restrict signal, float *restrict filter_kernel)
{
// Multiply the coefficients together and normalize the result














}



////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    runTest(argc, argv);
    return 0;
}


////////////////////////////////////////////////////////////////////////////////
//! Run a simple test for CUDA
////////////////////////////////////////////////////////////////////////////////
void runTest(int argc, char **argv)
{
    // Initalize GPU (set device ID)
    initGPU(argc, argv);

    // CUDA events for timing
    cudaEvent_t start, stop, startPointwise, stopPointwise;
    checkCudaErrors(cudaEventCreate(&start));
    checkCudaErrors(cudaEventCreate(&stop));
    checkCudaErrors(cudaEventCreate(&startPointwise));
    checkCudaErrors(cudaEventCreate(&stopPointwise));
    
    // Allocate host memory for the signal and filter
    Complex *h_signal = (Complex *)malloc(sizeof(Complex) * SIGNAL_SIZE);
    Complex *h_filter_kernel = (Complex *)malloc(sizeof(Complex) * FILTER_KERNEL_SIZE);
    
    // Initalize the memory for the signal
    for (unsigned int i = 0; i < SIGNAL_SIZE; ++i) {
        h_signal[i].x = rand() / (float)RAND_MAX;
        h_signal[i].y = 0;
    }

    // Initalize the memory for the filter
    for (unsigned int i = 0; i < FILTER_KERNEL_SIZE; ++i) {
        h_filter_kernel[i].x = rand() / (float)RAND_MAX;
        h_filter_kernel[i].y = 0;
    }

    // Pad signal and filter kernel
    Complex *h_padded_signal;
    Complex *h_padded_filter_kernel;
    int new_size = PadData(h_signal, &h_padded_signal, SIGNAL_SIZE,
                           h_filter_kernel, &h_padded_filter_kernel, FILTER_KERNEL_SIZE);
    int mem_size = sizeof(Complex) * new_size;

    // Allocate device memory for signal
    Complex *d_signal;
    checkCudaErrors(cudaMalloc((void **)&d_signal, mem_size));
    // Copy host memory to device
    checkCudaErrors(cudaMemcpy(d_signal, h_padded_signal, mem_size,
                               cudaMemcpyHostToDevice));

    // Allocate device memory for filter kernel
    Complex *d_filter_kernel;
    checkCudaErrors(cudaMalloc((void **)&d_filter_kernel, mem_size));
    
    // Create CUFFT plan
    cufftHandle plan;
    cufftResult error = cufftPlan1d(&plan, new_size, CUFFT_C2C, 1);
    if (error != CUFFT_SUCCESS)
        printf("CUFFT Error: %d", error);

    // Start timing
    checkCudaErrors(cudaEventRecord(start, 0));

    // Copy host memory to device
    checkCudaErrors(cudaMemcpy(d_filter_kernel, h_padded_filter_kernel, mem_size,
                             cudaMemcpyHostToDevice));

    // Transform signal and kernel
    printf("Transforming signal cufftExecC2C\n");
    error = cufftExecC2C(plan, (cufftComplex *)d_signal, (cufftComplex *)d_signal, CUFFT_FORWARD);
    error = cufftExecC2C(plan, (cufftComplex *)d_filter_kernel, (cufftComplex *)d_filter_kernel, CUFFT_FORWARD);
    if (error != CUFFT_SUCCESS)
        printf("CUFFT Error: %d", error);

    // Multiply the coefficients together and normalize the result
    printf("Performing point-wise complex multiply and scale.\n");
    checkCudaErrors(cudaEventRecord(startPointwise, 0));
    complexPointwiseMulAndScale(new_size, (float *restrict) d_signal, (float *restrict)  d_filter_kernel);
    checkCudaErrors(cudaEventRecord(stopPointwise, 0));

    // Transform signal back
    printf("Transforming signal back cufftExecC2C\n");
    error = cufftExecC2C(plan, (cufftComplex *)d_signal, (cufftComplex *)d_signal, CUFFT_INVERSE);
    if (error != CUFFT_SUCCESS)
        printf("CUFFT Error: %d", error);

    // Copy device memory to host
    Complex *h_convolved_signal = h_padded_signal;
    checkCudaErrors(cudaMemcpy(h_convolved_signal, d_signal, mem_size,
                               cudaMemcpyDeviceToHost));
    checkCudaErrors(cudaEventRecord(stop, 0));
    
    // Allocate host memory for the convolution result
    Complex *h_convolved_signal_ref = (Complex *)malloc(sizeof(Complex) * SIGNAL_SIZE);

    // Convolve on the host
    printf("Performing Convolution on the host and checking correctness\n");
    Convolve(h_signal, SIGNAL_SIZE,
             h_filter_kernel, FILTER_KERNEL_SIZE,
             h_convolved_signal_ref);

    // check result
    int testResult = compareL2((float *)h_convolved_signal_ref, 
                               (float *)h_convolved_signal, 
                               2 * SIGNAL_SIZE, 1e-5f);

    float ms, msPointwise;
    cudaEventSynchronize(stop);
//    checkCudaErrors(cudaEventElapsedTime(&ms, start, stop));
//    checkCudaErrors(cudaEventElapsedTime(&msPointwise, startPointwise, stopPointwise));
//    printf("\nSignal size: %d, filter size: %d\n", SIGNAL_SIZE, FILTER_KERNEL_SIZE);
//    printf("Total Device Convolution Time: %f ms (%f for point-wise convolution)\n", ms, msPointwise);
    
    printf("Test %s\n", testResult == 1 ? "PASSED" : "FAILED");

    //Destroy CUFFT context
    error = cufftDestroy(plan);
    if (error != CUFFT_SUCCESS)
        printf("CUFFT Error: %d", error);

    // cleanup memory
    free(h_signal);
    free(h_filter_kernel);
    free(h_padded_signal);
    free(h_padded_filter_kernel);
    free(h_convolved_signal_ref);
    checkCudaErrors(cudaFree(d_signal));
    checkCudaErrors(cudaFree(d_filter_kernel));
    checkCudaErrors(cudaEventDestroy(start));
    checkCudaErrors(cudaEventDestroy(stop));
    //cutilDeviceReset();
}

// Initialize GPU stuff: set device id
void initGPU(int argc, char** argv) {
        // gets the device id (if specified) to run on
        int devId = -1;
        int devCount = 0;
        if (argc > 1) {
                devId = atoi(argv[1]);
                cudaGetDeviceCount(&devCount);
		//devCount = acc_get_num_devices(acc_device_nvidia);
                if (devId < 0 || devId >= devCount) {
                        printf("The specified device ID is not supported.\n");
                        exit(1);
                }
        }
        if (devId != -1) {
                cudaSetDevice(devId);
		//acc_set_device_num(devId,acc_device_nvidia);
        }
        // creates a context on the GPU just to
        // exclude initialization time from computations
        cudaFree(0);
	//acc_init(acc_device_nvidia);

        // print device id
        cudaGetDevice(&devId);
	//devId = acc_get_device_num(acc_device_nvidia);
        printf("Running on GPU with ID %d.\n\n", devId);
}



// Pad data
int PadData(const Complex *signal, Complex **padded_signal, int signal_size,
            const Complex *filter_kernel, Complex **padded_filter_kernel, int filter_kernel_size)
{
    int minRadius = filter_kernel_size / 2;
    int maxRadius = filter_kernel_size - minRadius;
    int new_size = signal_size + maxRadius;

    // Pad signal
    Complex *new_data = (Complex *)malloc(sizeof(Complex) * new_size);
    memcpy(new_data + 0, signal, signal_size * sizeof(Complex));
    memset(new_data + signal_size, 0, (new_size - signal_size) * sizeof(Complex));
    *padded_signal = new_data;

    // Pad filter
    new_data = (Complex *)malloc(sizeof(Complex) * new_size);
    memcpy(new_data + 0, filter_kernel + minRadius, maxRadius * sizeof(Complex));
    memset(new_data + maxRadius, 0, (new_size - filter_kernel_size) * sizeof(Complex));
    memcpy(new_data + new_size - minRadius, filter_kernel, minRadius * sizeof(Complex));
    *padded_filter_kernel = new_data;

    return new_size;
}

